# coding: utf-8
from __future__ import absolute_import

### DEPS
import unittest
import sys
import six
import yaml
import os.path as op
import warnings

from .test_ontology import TestProntoOntology
from .              import utils


# Make sure we're using the local pronto library
sys.path.insert(0, op.dirname(op.dirname(op.abspath(__file__))))
import pronto


### TESTS
class TestProntoOboFoundry(TestProntoOntology):

    @classmethod
    @utils.ciskip
    def register_tests(cls):
        """Register tests for each ontology of the obofoundry"""

        foundry_url = "http://www.obofoundry.org/registry/ontologies.yml"
        foundry_rq = six.moves.urllib.request.urlopen(foundry_url)
        foundry_yaml = yaml.load(foundry_rq)

        # products = ( o['products'] for o in foundry_yaml['ontologies'] if 'products' in o )
        # for product in products:
        #     cls.add_test(product)

        for o in foundry_yaml['ontologies']:
            if 'products' in o:
                for product in o['products']:
                    cls.add_test(product)

    @classmethod
    def add_test(cls, product):
        """Add test for each product found in the obofoundry yaml file"""

                             #CRASH       #INF. WAIT
        #if product["id"] in ("chebi.obo", "dideo.owl"):
        #    return

        url, name = product["ontology_purl"], product["id"]

        def _foundry_noimports(self):
            onto = pronto.Ontology(url, False)
            self.check_ontology(onto)

        def _foundry_imports(self):
            onto = pronto.Ontology(url)
            self.check_ontology(onto)

        setattr(cls, "test_{}_foundry_noimports".format(name), _foundry_noimports)
        setattr(cls, "test_{}_foundry_imports".format(name),   _foundry_imports)


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    TestProntoOboFoundry.register_tests()
    suite.addTests(loader.loadTestsFromTestCase(TestProntoOboFoundry))
    return suite


def setUpModule():
    warnings.simplefilter('ignore')

def tearDownModule():
    warnings.simplefilter(warnings.defaultaction)



import os
import sys

try:
    from unittest import mock
except ImportError:
    import mock

# if os.getcwd().endswith("pronto"):
#     os.chdir("tests")

def ciskip(func):
    """Don't do anything if in CI environment
    """
    if "CI" in os.environ and os.environ["CI"].lower()=="true":
        def _pass(*args, **kwargs):
            pass
        return _pass
    else:
        return func

def py2skip(func):
    """Don't do anything if python version is not 3
    """
    if sys.version_info[0]==3:
        return func
    else:
        def _pass(*args, **kwargs):
            pass
        return _pass



"""
pronto.parser.owl
=================

This module defines the Owl parsing method.
"""
from __future__ import unicode_literals

import itertools
import collections
import six
#FEAT# DESERIALIZE AS DATES
#FEAT# import dateutil.parser

try:
    import lxml.etree as etree
except ImportError:
    try:
        import xml.etree.cElementTree as etree
    except ImportError:
        import xml.etree.ElementTree as etree

from .              import Parser
from .utils         import owl_ns, owl_to_obo, OwlSection
from ..relationship import Relationship
from ..term         import Term


RDF_ABOUT = "{{{}}}{}".format(owl_ns['rdf'], 'about')
RDF_RESOURCE = "{{{}}}{}".format(owl_ns['rdf'], 'resource')
RDF_DATATYPE = "{{{}}}{}".format(owl_ns['rdf'], 'datatype')
OWL_CLASS = "{{{}}}{}".format(owl_ns['owl'], 'Class')
OWL_ONTOLOGY = "{{{}}}{}".format(owl_ns['owl'], 'Ontology')



class OwlXMLParser(Parser):
    """Abstract OwlXMLParser.

    Provides functions common to all OwlXMLParsers, such as a function to
    extract ontology terms id from a url, or the common :obj:`hook` method.
    """

    ns = owl_ns
    extensions = ('.owl', '.ont', '.owl.gz', '.ont.gz')

    @classmethod
    def hook(self, **kwargs):
        """Returns True if this parser should be used.

        The current behaviour relies on filenames and file extension
        (.owl, .ont), but this is subject to change.
        """
        if 'force' in kwargs and kwargs['force']:
            return True
        if 'path' in kwargs:
            return kwargs['path'].endswith(self.extensions)

    @classmethod
    def parse(self, stream):
        """Parse the stream.

        This method is a classmethod, so it can be used to simply extract
        metadata, terms and imports of a file-like object without creating
        an ontology.

        Example:
            >>> fromp

        Parameters:
            stream (file handle): a binary stream of the file to parse

        Returns:
            dict: a dictionary containing the metadata headers
            dict: a dictionnary containing the terms
            set:  a set containing the imports
        """
        raise NotImplementedError

    @staticmethod
    def _get_id_from_url(url):
        if '#' in url:
            _id = url.split('#')[-1]
        else:
            _id = url.split('/')[-1]
        return _id.replace('_', ':')


class OwlXMLTreeParser(OwlXMLParser):

    @classmethod
    def parse(cls, stream):

        parser = etree.XMLParser()

        while True:
            chunk = stream.read(1024)
            if not chunk:
                break
            parser.feed(chunk)

        tree = parser.close()
        del parser

        meta, imports = cls._parse_meta(tree)
        _rawterms = cls._parse_terms(tree)
        del tree

        terms = cls._classify(_rawterms)
        del _rawterms
        meta = cls._relabel_owl_metadata(meta)

        return meta, terms, imports

    @staticmethod
    def _parse_meta(tree):

        imports = set()
        meta = collections.defaultdict(list)

        # tag.iter() starts on the element itself so we drop that
        for elem in itertools.islice(tree.find(OWL_ONTOLOGY).iter(), 1, None):

            basename = elem.tag.split('}', 1)[-1]
            if basename == 'imports':
                imports.add(next(six.itervalues(elem.attrib)))
            elif elem.text:
                meta[basename].append(elem.text)
            elif elem.get(RDF_RESOURCE) is not None:
                meta[basename].append(elem.get(RDF_RESOURCE))

        meta['import'] = list(imports)
        return meta, imports

    @classmethod
    def _parse_terms(cls, tree):

        _rawterms = []

        for rawterm in tree.iterfind(OWL_CLASS):

            if rawterm.get(RDF_ABOUT) is None:   # This avoids parsing a class
                continue                         # created by restriction

            _rawterms.append(collections.defaultdict(list))
            _rawterms[-1]['id'].append(cls._get_id_from_url(rawterm.get(RDF_ABOUT)))

            for elem in itertools.islice(rawterm.iter(), 1, None):

                basename = elem.tag.split('}', 1)[-1]
                if elem.text:
                    _rawterms[-1][basename].append(elem.text)
                elif elem.get(RDF_RESOURCE) is not None:
                    _rawterms[-1][basename].append(elem.get(RDF_RESOURCE))

        return _rawterms

    @classmethod
    def _classify(cls, _rawterms):
        terms = {}
        for rawterm in _rawterms:
            _id = cls._extract_obo_id(rawterm)
            name = cls._extract_obo_name(rawterm)
            desc = cls._extract_obo_desc(rawterm)
            relations = cls._extract_obo_relation(rawterm)
            others = cls._relabel_owl_properties(rawterm)
            terms[_id] = Term(_id, name, desc, dict(relations), others)
        return terms

    @staticmethod
    def _extract_obo_id(rawterm):
        _id = rawterm['id'][0]
        del rawterm['id']
        return _id

    @staticmethod
    def _extract_obo_name(rawterm):
        try:
            name = rawterm['label'][0]
        except IndexError:
            name = ''
        finally:
            del rawterm['label']
        return name

    @staticmethod
    def _extract_obo_desc(rawterm):
        desc = ''
        try:
            desc = rawterm['definition'][0]
        except IndexError:
            try:
                desc = rawterm['IAO_0000115'][0]
            except IndexError:
                pass
            finally:
                del rawterm['IAO_0000115']
        finally:
            del rawterm['definition']
        return desc

    @classmethod
    def _extract_obo_relation(cls, rawterm):
        relations = collections.defaultdict(list)

        for other in rawterm['subClassOf']:
            relations[Relationship('is_a')].append(
                cls._get_id_from_url(other)
            )
        del rawterm['subClassOf']

        return relations

    @staticmethod
    def _relabel_owl_properties(rawterm):
        new_term = {}
        for old_k, old_v in six.iteritems(rawterm):
            try:
                new_term[owl_to_obo[old_k]] = old_v
            except KeyError:
                new_term[old_k] = old_v
        return new_term

    @staticmethod
    def _relabel_owl_metadata(meta):
        new_meta = {}
        for old_k, old_v in six.iteritems(meta):
            try:
                new_meta[owl_to_obo[old_k]] = old_v
            except KeyError:
                new_meta[old_k] = old_v
        del meta
        return new_meta

OwlXMLTreeParser()


class _OwlXMLTarget(object):

    def __init__(self, meta=None, rawterms=None):
        self.ontology_tag = meta or collections.defaultdict(dict)
        self.classes = rawterms or []

        self.current_section = None
        self.current_tag = {'name':''}
        self.current_depth = 0

    def start(self, tag, attrib):
        self.current_depth += 1
        self.current_tag['name'] = tag

        if tag == OWL_ONTOLOGY and RDF_ABOUT in attrib:
            self.current_section = OwlSection.ontology
            self.ontology_tag['href'] = attrib[RDF_ABOUT]

        elif tag == OWL_CLASS:
            if RDF_ABOUT in attrib:
                self.current_section = OwlSection.classes
                self.classes.append(collections.defaultdict(dict))
                self.classes[-1]['id'] = {
                    'data': [OwlXMLParser._get_id_from_url(attrib[RDF_ABOUT])],
                }

        elif self.current_section == OwlSection.ontology:
            basename = self._get_basename(tag)
            try:
                self.ontology_tag[basename]['data'] = [attrib[RDF_RESOURCE]]
            except KeyError:
                pass
            try:
                self.ontology_tag[basename]['datatype'] = attrib[RDF_DATATYPE]
            except KeyError:
                self.ontology_tag[basename]['datatype'] = ''

        elif self.current_section == OwlSection.classes:
            basename = self._get_basename(tag)
            try:
                self.classes[-1][basename] = {
                    'data': [attrib[RDF_RESOURCE]],
                }
            except KeyError:
                pass
            try:
                self.classes[-1][basename]['datatype'] = attrib[RDF_DATATYPE]
            except KeyError:
                pass

    def end(self, tag):
        self.current_depth -= 1

        if tag == OWL_ONTOLOGY:
            self.current_section = None

        if tag == OWL_CLASS:
            if self.current_depth > 1:
                self.current_section = OwlSection.classes
            else:
                self.current_section = None

    def data(self, data):
        data = data.strip()

        if data:

            if self.current_section == OwlSection.ontology:
                basename = self._get_basename(self.current_tag['name'])
                try:
                    self.ontology_tag[basename]['data'].append(data)
                except KeyError:
                    self.ontology_tag[basename]['data'] = [data]

            elif self.current_section == OwlSection.classes:
                basename = self._get_basename(self.current_tag['name'])
                if basename in self.classes[-1]:
                    if 'data' in self.classes[-1][basename]:
                        self.classes[-1][basename]['data'].append(data)
                    else:
                        self.classes[-1][basename]['data'] = [data]

            del data

    # def comment(self, text):
    #     pass

    def close(self):
        return self.ontology_tag, self.classes

    @staticmethod
    def _get_basename(tag):
        return tag.split('}', 1)[-1]

class OwlXMLTargetParser(OwlXMLParser):

    @classmethod
    def parse(cls, stream):

        parser = etree.XMLParser(target=_OwlXMLTarget())

        while True:
            chunk = stream.read(1024)
            if not chunk:
                break
            parser.feed(chunk)

        meta, _rawterms = parser.close()
        del parser

        meta = cls._relabel_owl_metadata(meta)
        terms = cls._classify(_rawterms)
        del _rawterms

        try:
            imports = set(meta['imports'])
        except KeyError:
            imports = set()

        return meta, terms, imports

    @staticmethod
    def _relabel_owl_metadata(meta):

        new_meta = {}

        for k,v in meta.items():

            try:
                if v['datatype'] == "{}string".format(owl_ns['xsd']) and k not in {'hasDbXref', 'subClassOf'}:

                    try:
                        new_meta[owl_to_obo[k]] = ''.join(meta[k]['data'])
                    except KeyError:
                        new_meta[k] = ''.join(meta[k]['data'])

                else:
                    try:
                        new_meta[owl_to_obo[k]] = meta[k]['data']
                    except KeyError:
                        new_meta[k] = meta[k]['data']

                #FEAT# DESERIALIZE AS DATES
                #FEAT# elif v['datatype'] == "{xsd}dateTime".format_map(owl_ns):
                #FEAT#     self.ontology_tag[k]['data'] = dateutil.parser.parse(self.ontology_tag[k]['data'][0])

            except TypeError:
                pass

        return new_meta

    @classmethod
    def _classify(cls, rawterms):

        terms = {}

        #while True:
        for rawterm in rawterms:

            new_term = {}

            for k,v in rawterm.items():

                if not v:
                    continue

                try:

                    # if 'datatype' in v and v['datatype'] == "{}string".format(owl_ns['xsd']) and k != "id":

                    #     try:
                    #         new_term[owl_to_obo[k]] = ''.join(rawterm[k]['data'])
                    #     except KeyError:
                    #         new_term[k] = ''.join(rawterm[k]['data'])

                    if k == "subClassOf":
                        new_term[Relationship('is_a')] = [cls._get_id_from_url(t) for t in rawterm[k]['data']]

                    else:
                        try:
                            new_term[owl_to_obo[k]] = rawterm[k]['data']
                        except KeyError:
                            new_term[k] = rawterm[k]['data']

                except TypeError:
                    pass

            del rawterm

            _id = new_term['id'][0]
            del new_term['id']

            try:
                name = new_term['label'][0]
                del new_term['label']
            except KeyError:
                name = ''


            try:
                desc = ''.join(new_term['IAO_0000115'])
                del new_term['IAO_0000115']
            except KeyError:
                desc = ''

            relations = {}
            try:
                relations[Relationship('is_a')] = new_term[Relationship('is_a')]
                del new_term[Relationship('is_a')]
            except KeyError:
                pass

            terms[_id] = Term(_id, name, desc, relations, new_term)
            del new_term

        return terms

OwlXMLTargetParser()



# coding: utf-8
"""
pronto.ontology
===============

This submodule contains the definition of the Ontology class.


Multiprocessing
---------------

Ontology parsing relies on multiprocessing, which means it
isn't possible to use Ontology parsing within daemons. Still,
once the parsing is done, it is possible to use Ontologies
within processes or threads as all object implemented in pronto
are pickable.


"""
from __future__ import unicode_literals

import json
import os
import warnings
import six
import gzip
import contextlib

from six.moves.urllib.error import URLError, HTTPError

from .                   import __version__
from .term               import Term, TermList
from .parser             import Parser
from .parser.owl         import etree as _etree
from .utils              import ProntoWarning, output_str
from .relationship       import Relationship


class Ontology(object):
    """An ontology.

    Ontologies inheriting from this class will be able to use the same API as
    providing they generated the expected structure in the :func:`_parse`
    method.

    Examples:
        Import an ontology from a remote location::

            >>> from pronto import Ontology
            >>> envo = Ontology("https://raw.githubusercontent.com/"
            ... "EnvironmentOntology/envo/master/src/envo/envo.obo")

        Merge two local ontologies and export the merge::

            >>> uo = Ontology("tests/resources/uo.obo", False)
            >>> cl = Ontology("tests/resources/cl.ont", False)
            >>> uo.merge(cl)
            >>> with open('tests/run/merge.obo', 'w') as f:
            ...     lines_count = f.write(uo.obo)

        Export an ontology with its dependencies embedded::

            >>> cl = Ontology("tests/resources/cl.ont")
            >>> with open('tests/run/cl.obo', 'w') as f:
            ...     lines_count = f.write(cl.obo)

        Use the parser argument to force usage a parser::

            >>> cl = Ontology("tests/resources/cl.ont",
            ...               parser='OwlXMLTargetParser')


    Todo:
        * Add a __repr__ method to Ontology
    """

    def __init__(self, path=None, imports=True, import_depth=-1, timeout=2, parser=None):
        """
        """
        self.path = path
        self.meta = {}
        self.terms = {}
        self.imports = ()
        self.__parsedby__ = None

        if path is not None:

            with self._get_handle(path, timeout) as handle:
                self.parse(handle, parser)

            if self.__parsedby__ == None:
                raise ValueError("Could not find a suitable parser to parse {}".format(path))

            self.adopt()

            self.resolve_imports(imports, import_depth, parser)
            self.reference()

    def __contains__(self, item):
        """Check if the ontology contains a term

        It is possible to check if an Ontology contains a Term
        using an id or a Term object.

        Raises:
            TypeError: if argument (or left operand) is
                neither a string nor a Term

        Example:
            >>> 'CL:0002404' in cl
            True
            >>> from pronto import Term
            >>> Term('TST:001', 'tst') in cl
            False

        """
        if isinstance(item, (str, six.text_type)):
            return item in self.terms
        elif isinstance(item, Term):
            return item.id in self.terms
        else:
            raise TypeError("'in <Ontology>' requires string or Term as left operand, not {}".format(type(item)))

    def __iter__(self):
        """Returns an iterator over the Terms of the Ontology

        For convenience of implementation, the returned object is actually
        a generator object that returns each term of the ontology, sorted in
        alphabetic order of their id.

        Example:
            >>> for k in uo:
            ...    if 'basepair' in k.name:
            ...       print(k)
            <UO:0000328: kilobasepair>
            <UO:0000329: megabasepair>
            <UO:0000330: gigabasepair>
        """
        terms_accessions = sorted(self.terms.keys())
        return (self.terms[i] for i in terms_accessions)

    def __getitem__(self, item):
        """Overloaded object.__getitem__

        Method was overloaded to allow accessing to any Term of the Ontology
        using the Python dictionary syntax.

        Example:
            >>> cl['CL:0002380']
            <CL:0002380: oospore>
            >>> cl['CL:0002380'].relations
            {Relationship(is_a): [<CL:0000605: fungal asexual spore>]}

        """
        return self.terms[item]

    def __len__(self):
        """Returns the number of terms in the Ontology.
        """
        return self.terms.__len__()

    def __getstate__(self):

        meta = frozenset( (k, frozenset(v)) for k,v in six.iteritems(self.meta) )
        imports = self.imports
        path = self.path
        terms = frozenset(term for term in self)
        return (meta, imports, path, terms)

    def __setstate__(self, state):

        self.meta = {k:list(v) for (k,v) in state[0] }
        self.imports = state[1]
        self.path = state[2]
        self.terms = {t.id:t for t in state[3]}
        self.reference()

    def parse(self, stream, parser=None):
        """Parse the given file using available Parser instances

        Raises:
            TypeError: when the parser argument is not a string or None,
            ValueError: when the parser argument is a string that does
                not name a Parser

        """

        if parser is None:
            FORCE = False
            parsers = Parser._instances.values()
        elif isinstance(parser, (str, six.text_type)):
            if parser in Parser._instances:
                FORCE = True
                parsers = [Parser._instances[parser]]
            else:
                raise ValueError("could not find parser: {}".format(parser))
        else:
            raise TypeError("parser must be {types} or None, not {actual}".format(
                types = " or ".join([six.text_type.__name__, six.binary_type.__name__]),
                actual=type(parser).__name__,
            ))

        for p in parsers:
            if p.hook(stream=stream, path=self.path, force=FORCE):
                self.meta, self.terms, self.imports = p.parse(stream)
                self.__parsedby__ = type(p).__name__
                break

    def adopt(self):
        """Make terms aware of their children via complementary relationships

        This is done automatically when using the :obj:`merge` and :obj:`include`
        methods as well as the :obj:`__init__` method, but it should be called in
        case of manual editing of the parents or children of a Term.
        """

        valid_relationships = set(Relationship._instances.keys())

        relationships = [
            (parent, relation.complement(), term.id)
                for term in six.itervalues(self.terms)
                    for relation in term.relations
                        for parent in term.relations[relation]
                        	if relation.complementary
                        		and relation.complementary in valid_relationships
        ]

        relationships.sort(key=lambda x: x[2])

        for parent, rel, child in relationships:

            if rel is None:
                break

            try:
                parent = parent.id
            except AttributeError:
                pass

            if parent in self.terms:
                try:
                    if not child in self.terms[parent].relations[rel]:
                        self.terms[parent].relations[rel].append(child)
                except KeyError:
                    self[parent].relations[rel] = [child]

        del relationships

    def reference(self):
        """Make relationships point to classes of ontology instead of ontology id

        This is done automatically when using the :obj:`merge` and :obj:`include`
        methods as well as the :obj:`__init__` method, but it should be called in
        case of manual changes of the relationships of a Term.
        """
        for termkey,termval in six.iteritems(self.terms):
            for relkey, relval in six.iteritems(termval.relations):
                temprel = TermList()
                for x in relval:
                    try:
                        temprel.append(self.terms[x])
                    except KeyError:
                        if isinstance(x, Term):
                            temprel.append(x)
                        else:
                            temprel.append(Term(x,'',''))
                self.terms[termkey].relations[relkey] = temprel
                del temprel
                del relval

            # relvalref = { relkey: TermList(
            #                         [self.terms[x] if x in self.terms
            #                             else Term(x, '', '')
            #                                 if not isinstance(x, Term)
            #                             else x for x in relval]
            #                    )
            #             for relkey, relval in six.iteritems(termval.relations) }
            # self.terms[termkey].relations.update(relvalref)

    def resolve_imports(self, imports, import_depth, parser=None):
        """Import required ontologies.
        """
        if imports and import_depth:
            for i in list(self.imports):
                try:

                    if os.path.exists(i) or i.startswith(('http', 'ftp')):
                        self.merge(Ontology(i, import_depth=import_depth-1, parser=parser))

                    else: # try to look at neighbouring ontologies
                        self.merge(Ontology( os.path.join(os.path.dirname(self.path), i),
                                             import_depth=import_depth-1, parser=parser))

                except (IOError, OSError, URLError, HTTPError, _etree.ParseError) as e:
                    warnings.warn("{} occured during import of "
                                  "{}".format(type(e).__name__, i),
                                  ProntoWarning)

    def include(self, *terms):
        """Add new terms to the current ontology.

        Raises:
            TypeError: when the arguments is (are) neither a TermList nor a Term.

        Note:
            This will also recursively include terms in the term's relations
            dictionnary, but it is considered bad practice to do so. If you
            want to create your own ontology, you should only add an ID (such
            as 'ONT:001') to your terms relations, and let the Ontology link
            terms with each other.

        Examples:
            Create a new ontology from scratch

            >>> from pronto import Term, Relationship
            >>> t1 = Term('ONT:001','my 1st term',
            ...           'this is my first term')
            >>> t2 = Term('ONT:002', 'my 2nd term',
            ...           'this is my second term',
            ...           {Relationship('part_of'): ['ONT:001']})
            >>> ont = Ontology()
            >>> ont.include(t1, t2)
            >>>
            >>> 'ONT:002' in ont
            True
            >>> ont['ONT:001'].children
            [<ONT:002: my 2nd term>]

        """
        ref_needed = False

        for term in terms:

            if isinstance(term, TermList):
                ref_needed = ref_needed or self._include_term_list(term)
            elif isinstance(term, Term):
                ref_needed = ref_needed or self._include_term(term)
            else:
                raise TypeError('include only accepts <Term> or <TermList> as arguments')

        self.adopt()
        self.reference()

    def merge(self, other):
        """Merges another ontology into the current one.

        Raises:
            TypeError: When argument is not an Ontology object.

        Example:
            >>> from pronto import Ontology
            >>> nmr = Ontology('http://nmrml.org/cv/v1.0.rc1/nmrCV.owl', False)
            >>> po = Ontology('https://raw.githubusercontent.com/Planteome'
            ... '/plant-ontology/master/po.obo', False)
            >>> 'NMR:1000271' in nmr
            True
            >>> 'NMR:1000271' in po
            False
            >>> po.merge(nmr)
            >>> 'NMR:1000271' in po
            True

        """
        if not isinstance(other, Ontology):
            raise TypeError("'merge' requires an Ontology as argument,"
                            " not {}".format(type(other)))

        self.terms.update(other.terms)
        self._empty_cache()
        self.adopt()
        self.reference()

    @staticmethod
    @contextlib.contextmanager
    def _get_handle(path, timeout=2):

        REMOTE = path.startswith(('http', 'ftp'))
        ZIPPED = path.endswith('gz')

        if REMOTE:
            req = six.moves.urllib.request.Request(path, headers={'HTTP_CONNECTION': 'keep-alive'})
            if not ZIPPED or (ZIPPED and six.PY3):
                handle = six.moves.urllib.request.urlopen(req, timeout=timeout)
                if ZIPPED:
                    handle = gzip.GzipFile(fileobj=handle)
            else:
                raise NotImplementedError("Cannot parse a remote zipped file (this is an urllib2 limitation)")

        else:
            if os.path.exists(path):
                handle = open(path, 'rb')
            else:
                raise OSError('Ontology file {} could not be found'.format(path))
            if ZIPPED:
                handle = gzip.GzipFile(fileobj=handle)

        try:
            yield handle
        finally:
            handle.close()

    def _include_term_list(self, termlist):
        """Add terms from a TermList to the ontology.
        """
        ref_needed = False
        for term in termlist:
            ref_needed = ref_needed or self._include_term(term)
        return ref_needed

    def _include_term(self, term):
        """Add a single term to the current ontology

        It is needed to dereference any term in the term's relationship
        and then to build the reference again to make sure the other
        terms referenced in the term's relations are the one contained
        in the ontology (to make sure changes to one term in the ontology
        will be applied to every other term related to that term).
        """
        ref_needed = False

        if term.relations:

            for k,v in six.iteritems(term.relations):
                for i,t in enumerate(v):

                    #if isinstance(t, Term):
                    try:

                        if not t.id in self:
                            self._include_term(t)

                        term.relations[k][i] = t.id

                    except AttributeError:
                        pass

                    ref_needed = True

        self.terms[term.id] = term
        return ref_needed

    def _empty_cache(self, termlist=None):
        """Empty associated cache of each Term object

        This method is called when merging Ontologies or including
        new terms in the Ontology to make sure the cache of each
        term is cleaned and avoid returning wrong memoized values
        (such as Term.rchildren() TermLists, which get memoized for
        performance concerns)
        """
        if termlist is None:
            for term in six.itervalues(self.terms):
                term._empty_cache()
        else:
            for term in termlist:
                try:
                    self.terms[term.id]._empty_cache()
                except AttributeError:
                    self.terms[term]._empty_cache()

    @output_str
    def _obo_meta(self):
        """Generates the obo metadata header

        Generated following specs of the official format guide:
        ftp://ftp.geneontology.org/pub/go/www/GO.format.obo-1_4.shtml

        Todo:
            * Change the `auto-generated-by` tag to be **auto-generated-by pronto pronto.__version__**

        """
        metatags = [
            "format-version", "data-version", "date", "saved-by",
            "auto-generated-by","import", "subsetdef", "synonymtypedef",
            "default-namespace", "namespace-id-rule", "idspace",
            "treat-xrefs-as-equivalent", "treat-xrefs-as-genus-differentia",
            "treat-xrefs-as-is_a", "remark", "ontology"
        ]

        obo_meta = "\n".join(

            [ # official obo tags
                "{}: {}".format(k, x)
                        for k in metatags[:-1]
                            if k in self.meta
                                for x in self.meta[k]
            ] + [ # eventual other metadata added to remarks
                "remark: {}: {}".format(k, x)
                    for k,v in sorted(six.iteritems(self.meta), key=lambda x: x[0])
                        for x in v
                            if k not in metatags
            ] +      ["ontology: {}".format(x) for x in self.meta["ontology"]]
                            if "ontology" in self.meta
                else ["ontology: {}".format(self.meta["namespace"][0].lower())]
                            if "namespace" in self.meta
                else []

        )

        return obo_meta

    @property
    def json(self):
        """Returns the ontology serialized in json format.

        Example:
            >>> j = uo.json
            >>> all(term.id in j for term in uo)
            True

        Note:
            It is possible to save and load an ontology to and from
            json format, although it is cleaner to save and load
            an ontology in Obo format (as the json export doesn't store
            metadata, only terms).

        """
        return json.dumps( self.terms, indent=4, sort_keys=True,
                          default=lambda o: o.__deref__)

    @property
    @output_str
    def obo(self):
        """Returns the ontology serialized in obo format.
        """
        meta = self._obo_meta()
        if not isinstance(meta, six.text_type):
            meta = meta.decode('utf-8')
        meta = [meta] if meta else []

        if six.PY2:
            try: # if 'namespace' in self.meta:
                return "\n\n".join( meta + [
                    t.obo.decode('utf-8')
                        for t in self
                            if t.id.startswith(self.meta['namespace'][0])
                ])
            except KeyError:
                return "\n\n".join( meta + [
                    t.obo.decode('utf-8') for t in self
                ])
        elif six.PY3:
            try: # if 'namespace' in self.meta:
                return "\n\n".join( meta + [
                    t.obo for t in self
                        if t.id.startswith(self.meta['namespace'][0])
                ])
            except KeyError:
                return "\n\n".join( meta + [
                    t.obo for t in self
                ])




# coding: utf-8
import sys
import six
import os
import time
import functools
import unittest
import importlib
import warnings

from .utils import mock

# Make sure we're using the local pronto library
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import pronto



class TestProntoParser(unittest.TestCase):

    def setUp(self):
        self.resources_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources")

    def _check(self, m,t,i, *args, **kwargs):
        #FEAT# Store internals as utf-8
        #self._check_utf8(m,t,i)
        if 'exp_len' in kwargs:
            self._check_len(t, kwargs['exp_len'])

    def _check_utf8(self, m, t, i):
        """Check that inner properties are stored as utf-8
        """
        for term in six.itervalues(t):
            # six.text_type is str in Py3, unicode in Py2
            self.assertIsInstance(term.id, six.text_type)
            self.assertIsInstance(term.name, six.text_type)
            self.assertIsInstance(term.desc, six.text_type)

        for importpath in i:
            self.assertIsInstance(importpath, six.text_type)

        for k,v in six.iteritems(m):
            self.assertIsInstance(k, six.text_type)
            for x in v:
                self.assertIsInstance(x, six.text_type)

    def _check_len(self, t, exp_len):
        """Check the terms dict length matches the expected length
        """
        self.assertEqual(len(t), exp_len)



class TestProntoOwlParser(TestProntoParser):

    def setUp(self):
        super(TestProntoOwlParser, self).setUp()
        self.lxml_etree = importlib.import_module("lxml.etree")
        self.xml_etree = importlib.import_module("xml.etree.ElementTree")
        self.cxml_etree = importlib.import_module("xml.etree.cElementTree")

    @staticmethod
    def _parse(parser, path):
        if path.startswith(('http', 'ftp')):
            handle = six.moves.urllib.request.urlopen(path)
        else:
            handle = open(path, 'rb')
        try:
            m,t,i = parser.parse(handle)
        finally:
            handle.close()
        return m,t,i

class TestProntoOwlTargetParser(TestProntoOwlParser):

    def test_with_lxml_etree(self):
        with mock.patch("pronto.parser.owl.etree", self.lxml_etree):
            m,t,i = self._parse(
                pronto.parser.owl.OwlXMLTargetParser(),
                os.path.join(self.resources_dir, 'cl.ont'),
            )
        self._check(m,t,i, exp_len=2348)

    def test_with_xml_elementTree(self):
        with mock.patch("pronto.parser.owl.etree", self.cxml_etree):
            m,t,i = self._parse(
                pronto.parser.owl.OwlXMLTargetParser(),
                os.path.join(self.resources_dir, 'cl.ont'),
            )
        self._check(m,t,i, exp_len=2348)

    def test_with_xml_cElementTree(self):
        with mock.patch("pronto.parser.owl.etree", self.xml_etree):
            m,t,i = self._parse(
                pronto.parser.owl.OwlXMLTargetParser(),
                os.path.join(self.resources_dir, 'cl.ont'),
            )
        self._check(m,t,i, exp_len=2348)

    def test_with_lxml_etree_remote(self):
        with mock.patch("pronto.parser.owl.etree", self.lxml_etree):
            m,t,i = self._parse(
                pronto.parser.owl.OwlXMLTargetParser(),
                "http://nmrml.org/cv/v1.0.rc1/nmrCV.owl",
            )
        self._check(m,t,i, exp_len=685)

    def test_with_xml_elementTree_remote(self):
        with mock.patch("pronto.parser.owl.etree", self.cxml_etree):
            m,t,i = self._parse(
                pronto.parser.owl.OwlXMLTargetParser(),
                "http://nmrml.org/cv/v1.0.rc1/nmrCV.owl",
            )
        self._check(m,t,i, exp_len=685)

    def test_with_xml_cElementTree_remote(self):
        with mock.patch("pronto.parser.owl.etree", self.xml_etree):
            m,t,i =  self._parse(
                pronto.parser.owl.OwlXMLTargetParser(),
                "http://nmrml.org/cv/v1.0.rc1/nmrCV.owl",

            )
        self._check(m,t,i, exp_len=685)

class TestProntoOwlTreeParser(TestProntoOwlParser):

    def test_with_lxml_etree(self):
        with mock.patch("pronto.parser.owl.etree", self.lxml_etree):
            m,t,i = self._parse(
                pronto.parser.owl.OwlXMLTreeParser(),
                os.path.join(self.resources_dir, 'cl.ont'),
            )
        self._check(m,t,i, exp_len=2348)

    def test_with_xml_elementTree(self):
        with mock.patch("pronto.parser.owl.etree", self.cxml_etree):
            m,t,i = self._parse(
                pronto.parser.owl.OwlXMLTreeParser(),
                os.path.join(self.resources_dir, 'cl.ont'),
            )
        self._check(m,t,i, exp_len=2348)

    def test_with_xml_cElementTree(self):
        with mock.patch("pronto.parser.owl.etree", self.xml_etree):
            m,t,i = self._parse(
                pronto.parser.owl.OwlXMLTreeParser(),
                os.path.join(self.resources_dir, 'cl.ont'),
            )
        self._check(m,t,i, exp_len=2348)

    def test_with_lxml_etree_remote(self):
        with mock.patch("pronto.parser.owl.etree", self.lxml_etree):
            m,t,i = self._parse(
                pronto.parser.owl.OwlXMLTreeParser(),
                "http://nmrml.org/cv/v1.0.rc1/nmrCV.owl",
            )
        self._check(m,t,i, exp_len=685)

    def test_with_xml_elementTree_remote(self):
        with mock.patch("pronto.parser.owl.etree", self.cxml_etree):
            m,t,i = self._parse(
                pronto.parser.owl.OwlXMLTreeParser(),
                "http://nmrml.org/cv/v1.0.rc1/nmrCV.owl",
            )
        self._check(m,t,i, exp_len=685)

    def test_with_xml_cElementTree_remote(self):
        with mock.patch("pronto.parser.owl.etree", self.xml_etree):
            m,t,i = self._parse(
                pronto.parser.owl.OwlXMLTreeParser(),
                "http://nmrml.org/cv/v1.0.rc1/nmrCV.owl",
            )
        self._check(m,t,i, exp_len=685)


def setUpModule():
    warnings.simplefilter('ignore')

def tearDownModule():
    warnings.simplefilter(warnings.defaultaction)



# coding: utf-8
"""
pronto.term
===========

This module defines the Term and TermList classes.
"""
from __future__ import unicode_literals

import six

from .relationship import Relationship
from .utils import output_str

class Term(object):
    """An ontology term.

    Example:

        >>> from pronto import *
        >>> new_term = Term('TR:001', 'new term', 'a new term')
        >>> linked_term = Term('TR:002', 'other new', 'another term',
        ...                    { Relationship('is_a'): 'TR:001'})

        >>> ms = Ontology('https://raw.githubusercontent.com/'
        ...               'HUPO-PSI/psi-ms-CV/master/psi-ms.obo')
        >>> type(ms['MS:1000015'])
        <class 'pronto.term.Term'>

    """

    def __init__(self, tid, name='', desc='', relations=None, other=None):
        """

        Parameters:
            tid (str): the Term id (e.g. MS:1000031)
            name (str): the name of the Term in human language
            desc (str): a description of the Term
            relations (dict): a dictionary containing the other
                terms the Term is in a relationship with.
            other (dict): other information about the term

        """
        self.id = tid
        self.name = name
        self.desc = desc
        self.relations = relations or {}
        self.other = other or {}

        self._rchildren  = {}
        self._rparents = {}
        self._children = None
        self._parents = None

    @output_str
    def __repr__(self):
        return "<{}: {}>".format(self.id, self.name)

    @property
    def parents(self):
        """The parents of the Term.

        Returns:
            :obj:`pronto.TermList`:
            a TermList containing all parents of the Term
            (other terms with which this Term has a "bottomup"
            relationship)

        Example:

            >>> for p in ms['MS:1000532'].parents: print(p.desc)
            "Thermo Finnigan software for data acquisition and analysis." [PSI:MS]
            "Acquisition software." [PSI:MS]
            "Analysis software." [PSI:MS]
            "Data processing software." [PSI:MS]

        """

        if self._parents is None:
            bottomups = tuple(Relationship.bottomup())

            self._parents = TermList()
            self._parents.extend(
                [ other
                    for rship,others in six.iteritems(self.relations)
                        for other in others
                            if rship in bottomups
                ]

            )

        return self._parents

    @property
    def children(self):
        """The children of the Term.

        Returns:
            :obj:`pronto.TermList`:

            a TermList containing all parents of the Term
            (other terms with which this Term has a "topdown"
            relationship)

        Example:

            >>> ms['MS:1000452'].children
            [<MS:1000530: file format conversion>, <MS:1000543: data processing action>]

        """

        if self._children is None:

            topdowns = tuple(Relationship.topdown())
            self._children = TermList()
            self._children.extend(
                [ other
                    for rship,others in six.iteritems(self.relations)
                        for other in others
                            if rship in topdowns
                ]
            )

        return self._children

    @property
    @output_str
    def obo(self):
        """The Term serialized in an Obo Term stanza.

        Example:

            >>> print(ms['MS:1000031'].obo)
            [Term]
            id: MS:1000031
            name: instrument model
            def: "Instrument model name not including the vendor's name." [PSI:MS]
            relationship: part_of MS:1000463 ! instrument


        .. note::
            The following guide was used:
            ftp://ftp.geneontology.org/pub/go/www/GO.format.obo-1_4.shtml

        """

        metatags = ["id", "is_anonymous", "name", "namespace","alt_id", "def","comment",
                    "subset","synonym","xref","builtin","property_value","is_a",
                    "intersection_of","union_of","equivalent_to","disjoint_from",
                    "relationship","created_by","creation_date","is_obsolete",
                    "replaced_by", "consider"]

        obo = "\n".join(
            [

                "[Term]\nid: {}\nname: {}".format(self.id, self.name if self.name is not None else "")

            ] + [ #metatags from namespace to alt_id

                "{}: {}".format(k, self.other[k])
                    for k in metatags[3:6]
                        if k in self.other

            ] + ([ " ".join(["def:", self.desc]) ] if self.desc else [])


              + [ # metatags from comment to property_value

                  "{}: {}".format(k, self.other[k])
                        for k in metatags[7:12]
                            if k in self.other

            ] + [ #is_a

                "is_a: {} ! {}".format(companion.id, companion.name)
                    for relation in self.relations
                        if relation is Relationship('is_a')
                            for companion in self.relations[Relationship('is_a')]


            ] + [ #metatags from intersection_of to disjoint_from

                "{}: {}".format(k, self.other[k])
                    for k in metatags[13:17]
                        if k in self.other

            ] + [ #relationships

                "relationship: {} {} ! {}".format(relation.obo_name, companion.id, companion.name)
                    for relation in self.relations
                        if relation.direction=="bottomup" and relation is not Relationship('is_a')
                            for companion in self.relations[relation]

                    #for relation in Relationship.bottomup()
                    #    if relation in self.relations and relation is not Relationship('is_a')
                    #        for companion in self.relations[relation]

            ] + [ #metatags from created_by to consider

                "{}: {}".format(k, self.other[k])
                    for k in metatags[18:]
                        if k in self.other

            ]


        )

        return obo




        # obo =  "".join([ '[Term]', '\n',
        #                  'id: ', self.id, '\n',
        #                  'name: ', self.name if self.name is not None else '', '\n'])
        # if self.desc:
        #     obo = "".join([obo, 'def: ', self.desc, '\n'])

        # # add more bits of information
        # for k,v in six.iteritems(self.other):
        #     if isinstance(v, list):
        #         obo = "".join( [obo] + ["{}: {}\n".format(k, x) for x in v] )
        #     #    for x in v:
        #     #        obo = "".join([obo, k, ': ', x, '\n'])
        #     else:
        #         obo = "".join([obo,k, ': ', v, '\n'])
        #     #obo = "".join( [obo] + ["{}: {}\n".format(k, x) for x in v] )

        # # add relationships (only bottom up ones)

        # for relation in Relationship.bottomup():
        #     try:
        #         for companion in self.relations[relation]:

        #             if relation is not Relationship('is_a'):
        #                 obo = "".join([obo, 'relationship: '])
        #             obo = "".join([obo, relation.obo_name, ': '])

        #             try:
        #                 obo = "".join([obo, companion.id, ' ! ', companion.name, '\n'])
        #             except AttributeError:
        #                 obo = "".join([obo,companion, '\n'])

        #     except KeyError:
        #         continue

        # return obo.rstrip()

    @property
    def __deref__(self):
        """A dereferenced Term.__dict__.

        It only contains other Terms id to avoid circular references when
        creating a json.
        """
        return {
            'id': self.id,
            'name': self.name,
            'other': self.other,
            'desc': self.desc,
            'relations': {k.obo_name:v.id for k,v in six.iteritems(self.relations)}
         }

    def __getstate__(self):

        return (
            self.id,
            self.name,
            tuple((k,v) for k,v in six.iteritems(self.other)),
            self.desc,
            tuple((k.obo_name,v.id) for k,v in six.iteritems(self.relations)),
        )

    def __setstate__(self, state):

        self.id = state[0]
        self.name = state[1]
        self.other = {k:v for (k,v) in state[2]}
        self.desc = state[3]
        self.relations = {Relationship(k):v for k,v in state[4]}

    def _empty_cache(self):
        """
        Empties the cache of the Term's memoized functions.
        """
        self._children, self._parents = None, None
        self._rchildren, self._rparents = {}, {}

    def rchildren(self, level=-1, intermediate=True):
        """Create a recursive list of children.

        Parameters:
            level (int): The depth level to continue fetching children from
                (default is -1, to get children to the utter depths)
            intermediate (bool): Also include the intermediate children
                (default is True)

        Returns:
            :obj:`pronto.TermList`:
            The recursive children of the Term following the parameters

        """
        try:
            return self._rchildren[(level, intermediate)]

        except KeyError:

            rchildren = []

            if self.children and level:

                if intermediate or level==1:
                    rchildren.extend(self.children)

                for child in self.children:
                    rchildren.extend(child.rchildren(level=level-1,
                                                     intermediate=intermediate))

            rchildren = TermList(set(rchildren))
            self._rchildren[(level, intermediate)] = rchildren
            return rchildren

    def rparents(self, level=-1, intermediate=True):
        """Create a recursive list of children.

        Note that the :param:`intermediate` can be used to include every
        parents to the returned list, not only the most nested ones.

        Parameters:
            level (int): The depth level to continue fetching parents from
                (default is -1, to get parents to the utter depths)
            intermediate (bool): Also include the intermediate parents
                (default is True)

        Returns:
            :obj:`pronto.TermList`:
            The recursive children of the Term following the parameters

        """
        try:
            return self._rparents[(level, intermediate)]

        except KeyError:

            rparents = []

            if self.parents and level:

                if intermediate or level==1:
                    rparents.extend(self.parents)

                for parent in self.parents:
                    rparents.extend(parent.rparents(level=level-1,
                                                     intermediate=intermediate))

            rparents = TermList(set(rparents))
            self._rparents[(level, intermediate)] = rparents
            return rparents


class TermList(list):
    """A list of Terms.

    TermList behaves exactly like a list, except it contains shortcuts to
    generate lists of terms' attributes.

    Example:
        >>> from pronto import *
        >>> nmr = Ontology('http://nmrml.org/cv/v1.0.rc1/nmrCV.owl')
        >>> type(nmr['NMR:1000031'].children)
        <class 'pronto.term.TermList'>

        >>> nmr['NMR:1000031'].children.id
        [u'NMR:1000122', u'NMR:1000156', u'NMR:1000157', u'NMR:1000489']
        >>> nmr['NMR:1400014'].relations[Relationship('is_a')]
        [<NMR:1400011: cardinal part of NMR instrument>]


    .. tip::
        It is also possible to call Term methods on a TermList to
        create another TermList::

            >>> nmr['NMR:1000031'].rchildren(3, False).rparents(3, False).id
            [u'NMR:1000031']

    """

    def __init__(self, elements=None):
        """
        """
        super(TermList, self).__init__(elements or [])
        self._check_content()


    def append(self, element):
        try:
            if not element in self:
                super(TermList, self).append(element)
        except TypeError:
            raise

    def _check_content(self):
        try:
            [ term.id for term in self ]
        except AttributeError:
            raise TypeError('TermList can only contain Terms.')

    def rparents(self, level=-1, intermediate=True):
        return TermList(set(
            [y for x in self for y in x.rparents(level, intermediate)]
        ))

    def rchildren(self, level=-1, intermediate=True):
        return TermList(set(
            [y for x in self for y in x.rchildren(level, intermediate)]
        ))

    @property
    def children(self):
        return TermList( [ y for x in self for y in x.children] )

    @property
    def parents(self):
        return TermList( [ y for x in self for y in x.parents] )

    @property
    def id(self):
        """Return a list containing id of all terms in current TermList
        """
        return [x.id for x in self]

    @property
    def name(self):
        return [x.name for x in self]

    @property
    def desc(self):
        return [x.desc for x in self]

    @property
    def other(self):
        return [x.other for x in self]

    @property
    def obo(self):
        return [x.obo for x in self]

    def __getstate__(self):
        return (x for x in self)

    def __setstate__(self, state):
        self.extend(state)

    def __contains__(self, term):
        """Check if the TermList contains a term.

        The method allows to check for the presence of a Term in a
        TermList based on a Term object or on a term accession number.

        Example:

            >>> from pronto import *
            >>> nmr = Ontology('http://nmrml.org/cv/v1.0.rc1/nmrCV.owl')
            >>> 'NMR:1000122' in nmr['NMR:1000031'].children
            True
            >>> nmr['NMR:1000122'] in nmr['NMR:1000031'].children
            True

        """
        try:
            _id = term.id
        except AttributeError:
            _id = term

        return any((t.id==_id if isinstance(t, Term) else t==_id for t in self))



#!/usr/bin/env python
# released under the GNU General Public License version 3.0 (GPLv3)

from setuptools import setup, find_packages
import warnings

warnings.simplefilter("ignore")
import pronto


def format_for_setup(requirement_file):
    """Build a list of requirements out of requirements.txt files.
    """
    requirements = []
    with open(requirement_file) as rq:
        for line in rq:
            line = line.strip()
            if line.startswith('-r'):
                other_requirement_file = line.split(' ', 1)[-1]
                requirements.extend(format_for_setup(other_requirement_file))
            elif line:
                requirements.append(line)
    return requirements

## SETUPTOOLS VERSION
setup(
    name='pronto',
    version=pronto.__version__,

    packages=find_packages(),

    py_modules=['pronto'],

    author= pronto.__author__,
    author_email= pronto.__author_email__,

    description="Python frontend to ontologies - a library to parse, create, browse and export ontologies.",
    long_description=open('README.rst').read(),

    run_requires= format_for_setup('requirements.txt'),
    test_requires = format_for_setup('requirements-test.txt'),

    extras_require = { extra:format_for_setup('requirements-{}.txt'.format(extra))
                        for extra in ['doc'] },

    include_package_data=True,

    url='http://github.com/althonos/pronto',

    test_suite="tests",

    classifiers=[
    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.2",
    "Programming Language :: Python :: 3.3",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: 3.5",
    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
    "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Operating System :: OS Independent",
    ],

    license="GPLv3",

    keywords = ['Bio-Informatics', 'Ontology', 'OBO', 'Owl', 'convert', 'parse'],

)

